"""
:copyright: (c) 2010-2015 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

import asyncio
import socket

from notifire import constants
from notifire.exceptions import APIError, RateLimited
from notifire.transport.base import AsyncTransport
from notifire.transport.http import HTTPTransport

try:
    import aiohttp
    has_aiohttp = True
except ImportError:
    has_aiohttp = False


class AioHTTPTransportBase(AsyncTransport, HTTPTransport):
    def __init__(self, timeout=constants.TIMEOUT, verify_ssl=True,
                 keepalive=True, family=socket.AF_INET, loop=None):
        if not has_aiohttp:
            raise ImportError('AioHttpTransport requires aiohttp.')

        self._keepalive = keepalive
        self._family = family
        if loop is None:
            loop = asyncio.get_event_loop()

        self._loop = loop

        super().__init__(timeout, verify_ssl)

        if self.keepalive:
            self._client_session = self._client_session_factory()

        self._closing = False

    @property
    def keepalive(self):
        return self._keepalive

    @property
    def family(self):
        return self._family

    def _client_session_factory(self):
        connector = aiohttp.TCPConnector(verify_ssl=self.verify_ssl,
                                         family=self.family,
                                         loop=self._loop)
        return aiohttp.ClientSession(connector=connector,
                                     loop=self._loop)

    async def _do_send(self, url, data, headers, success_cb, failure_cb):
        if self.keepalive:
            session = self._client_session
        else:
            session = self._client_session_factory()

        resp = None

        try:
            resp = await session.post(
                url,
                data=data,
                compress=False,
                headers=headers,
                timeout=self.timeout
            )

            code = resp.status
            if not 200 <= code < 300:
                msg = await resp.text()
                if code == 429:
                    try:
                        retry_after = resp.headers.get('retry-after')
                        retry_after = int(retry_after)
                    except (ValueError, TypeError):
                        retry_after = 0
                    failure_cb(RateLimited(msg, retry_after))
                else:
                    failure_cb(APIError(msg, code))
            else:
                success_cb()
        except asyncio.CancelledError:  # pylint: disable=W0706
            # do not mute asyncio.CancelledError
            raise
        except Exception as exc:
            failure_cb(exc)
        finally:
            if resp:
                resp.release()
            if not self.keepalive:
                await session.close()

    def _async_send(self, url, data, headers, success_cb, failure_cb):
        raise NotImplementedError

    async def _close(self):
        raise NotImplementedError

    def async_send(self, url, data, headers, success_cb, failure_cb):
        if self._closing:
            failure_cb(RuntimeError(
                '{} is closed'.format(self.__class__.__name__)
            ))
            return

        self._async_send(url, data, headers, success_cb, failure_cb)

    async def _close_coro(self, timeout=None):
        try:
            await asyncio.wait_for(
                self._close(), timeout=timeout)
        except asyncio.TimeoutError:
            pass
        finally:
            if self.keepalive:
                await self._client_session.close()

    def close(self, timeout=None):
        if self._closing:
            async def dummy():
                pass

            return dummy()

        self._closing = True

        return self._close_coro(timeout=timeout)


class AioHTTPTransport(AioHTTPTransportBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._tasks = set()

    def _async_send(self, url, data, headers, success_cb, failure_cb):
        coro = self._do_send(url, data, headers, success_cb, failure_cb)

        task = asyncio.ensure_future(coro, loop=self._loop)
        self._tasks.add(task)
        task.add_done_callback(self._tasks.remove)

    async def _close(self):
        await asyncio.gather(
            *self._tasks,
            return_exceptions=True
        )

        assert len(self._tasks) == 0



import gzip
import io
import json
import sys

from notifire import constants
from notifire.exceptions import InvalidNotificationLevel, MessageSizeExceeded


class Message:
    levels = ('info', 'warning', 'error')

    @classmethod
    def get_payload(cls, body, level='info', text=None, url=None):
        if level.lower() not in cls.levels:
            raise InvalidNotificationLevel(
                'Notification level must be one of {}'
                .format(cls.levels))
        payload = cls.payload(body, level, text, url)
        payload = json.dumps(payload).encode('utf-8')

        if not cls.check_size(payload):
            raise MessageSizeExceeded(
                'Size of a message exceeded maximum of {} bytes'
                .format(constants.MESSAGE_SIZE))

        return cls.encode(payload)

    @staticmethod
    def payload(body, level, text, url):
        return {
            'level': level,
            'message': {
                'titleBody': body,
                'text': text,
                'url': url
            }
        }

    @staticmethod
    def encode(data):
        body = io.BytesIO()
        with gzip.GzipFile(fileobj=body, mode='w') as f:
            f.write(data)
        return body.getvalue()

    @staticmethod
    def decode(data):
        return gzip.GzipFile(fileobj=io.BytesIO(data)).read()

    @staticmethod
    def check_size(payload):
        return sys.getsizeof(payload) < constants.MESSAGE_SIZE



import pytest

from notifire.client import Client
from notifire.transport import base as uut


class TestTransport:
    client = Client('service_api_key', transport=uut.Transport,
                    raise_send_errors=True)

    def test_send(self):
        with pytest.raises(NotImplementedError):
            self.client.send_notification('foo')


class TestAsyncTransport:
    client = Client('service_api_key', transport=uut.AsyncTransport,
                    raise_send_errors=True)

    def test_async_send(self):
        with pytest.raises(NotImplementedError):
            self.client.send_notification('foo')



"""
:copyright: (c) 2010-2015 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

try:
    from gevent import monkey; monkey.patch_all()  # NOQA
except ImportError:
    import pytest
    pytest.skip(
        'Testing gevent transport requires gevent.',
        allow_module_level=True
    )

import time

import mock

from notifire.client import Client
from notifire.transport import gevent as uut


class TestGeventedHTTPTransport:
    client = Client('service_api_key', transport=uut.GeventedHTTPTransport)

    @mock.patch.object(uut.GeventedHTTPTransport, '_done')
    @mock.patch('notifire.transport.http.HTTPTransport.send')
    def test_does_send(self, send, done):
        self.client.send_notification('foo')
        time.sleep(0)
        assert send.call_count == 1
        time.sleep(0)
        assert done.call_count == 1



"""
:copyright: (c) 2010-2015 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

import os
from tempfile import mkstemp
import time

import mock

from notifire.client import Client
from notifire.message import Message
from notifire.transport import threaded as uut


class DummyThreadedScheme(uut.ThreadedHTTPTransport):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.events = []
        self.send_delay = 0

    def send_sync(self, url, data, headers, success_cb, failure_cb):
        # delay sending the message, to allow us to tests that the shutdown
        # hook waits correctly
        time.sleep(self.send_delay)

        self.events.append((url, data, headers, success_cb, failure_cb))


class LoggingThreadedScheme(uut.ThreadedHTTPTransport):
    def __init__(self, filename, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.filename = filename

    def send_sync(self, url, data, headers, success_cb, failure_cb):
        with open(self.filename, 'a') as log:
            log.write(
                '{0} {1}\n'.format(
                    os.getpid(),
                    data['message']['titleBody'])
            )


class TestThreadedHTTPTransport:
    url = 'localhost'
    message = Message()
    client = Client('service_api_key', transport=uut.ThreadedHTTPTransport)

    @mock.patch('notifire.transport.http.HTTPTransport.send')
    def test_does_send(self, send):
        self.client.send_notification('foo', level='info')

        time.sleep(0.1)

        assert send.call_count == 1

    def test_shutdown_waits_for_send(self):
        transport = DummyThreadedScheme()
        transport.send_delay = 0.5

        data = self.message.get_payload(
            'foo',
            level='info',
            text=None,
            url=None
        )
        transport.async_send(self.url, data, None, None, None)

        time.sleep(0.1)

        # this should wait for the message to be sent
        transport.get_worker().main_thread_terminated()

        assert len(transport.events) == 1

    def test_fork_spawns_anew(self):
        transport = DummyThreadedScheme()
        transport.send_delay = 0.5

        data = self.message.get_payload(
            'foo',
            level='info',
            text=None,
            url=None
        )

        pid = os.fork()
        if pid == 0:
            time.sleep(0.1)

            transport.async_send(self.url, data, None, None, None)

            # this should wait for the message to get sent
            transport.get_worker().main_thread_terminated()

            assert len(transport.events) == 1
            # Use os._exit here so that pytest will not get confused about
            # what the hell we're doing here.
            os._exit(0)
        else:
            os.waitpid(pid, 0)

    def test_fork_with_active_worker(self):
        """
        Test threaded transport when forking an active worker.
        Forking a process doesn't clone the worker thread - make sure
        logging from both processes still works.
        """
        event1 = self.message.payload(
            'parent',
            level='info',
            text=None,
            url=None
        )
        event2 = self.message.payload(
            'child',
            level='info',
            text=None,
            url=None
        )
        fd, filename = mkstemp()
        try:
            os.close(fd)
            transport = LoggingThreadedScheme(filename)

            # Log from the parent process - starts the worker thread
            transport.async_send(self.url, event1, None, None, None)
            childpid = os.fork()

            if childpid == 0:
                # Log from child process
                transport.async_send(self.url, event2, None, None, None)

                # Ensure threaded worker has finished
                transport.get_worker().stop()
                os._exit(0)

            # Wait for the child process to finish
            os.waitpid(childpid, 0)
            assert os.path.isfile(filename)

            # Ensure threaded worker has finished
            transport.get_worker().stop()

            with open(filename, 'r') as logfile:
                events = dict(x.strip().split() for x in logfile.readlines())

            # Check that parent and child both logged successfully
            assert events == {
                str(os.getpid()): 'parent',
                str(childpid): 'child'
            }
        finally:
            os.remove(filename)



from notifire.transport.aiohttp import AioHTTPTransport  # NOQA
from notifire.transport.gevent import GeventedHTTPTransport  # NOQA
from notifire.transport.http import HTTPTransport  # NOQA
from notifire.transport.requests import RequestsHTTPTransport  # NOQA
from notifire.transport.threaded import ThreadedHTTPTransport  # NOQA
from notifire.transport.threaded_requests import ThreadedRequestsHTTPTransport  # NOQA



"""
:copyright: (c) 2010-2015 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

import asyncio
import logging

import mock
import pytest

from notifire.transport.aiohttp import AioHTTPTransport
from tests.transport.test_aiohttp.utils import Logger

pytestmark = pytest.mark.asyncio


async def test_basic(fake_server, notifire_client, wait):
    server = await fake_server()

    client, transport = notifire_client(server, AioHTTPTransport)

    client.send_notification('foo', level='info')

    await wait(transport)

    assert server.hits[200] == 1


async def test_no_keepalive(fake_server, notifire_client, wait):
    transport = AioHTTPTransport(keepalive=False)
    assert not hasattr(transport, '_client_session')
    await transport.close()

    server = await fake_server()

    client, transport = notifire_client(server, AioHTTPTransport)
    transport._keepalive = False
    session = transport._client_session

    def _client_session_factory():
        return session

    with mock.patch(
        'notifire.transport.aiohttp.AioHTTPTransport._client_session_factory',
        side_effect=_client_session_factory,
    ):
        client.send_notification('foo', level='info')

        await wait(transport)

        assert session.closed

        assert server.hits[200] == 1


async def test_close_timeout(fake_server, notifire_client):
    server = await fake_server()
    server.slop_factor = 100

    client, transport = notifire_client(server, AioHTTPTransport)

    client.send_notification('foo', level='info')

    await transport.close(timeout=0)

    assert server.hits[200] == 0


async def test_rate_limit(fake_server, notifire_client, wait):
    server = await fake_server()
    server.side_effect['status'] = 429

    with Logger('notifire', level=logging.ERROR) as log:
        client, transport = notifire_client(server, AioHTTPTransport)

        client.send_notification('foo', level='error')

        await wait(transport)

        assert server.hits[429] == 1

    msg = 'Notifire api responded with an error: RateLimited()'
    assert log.msgs[0] == msg


async def test_status_500(fake_server, notifire_client, wait):
    server = await fake_server()
    server.side_effect['status'] = 500

    with Logger('notifire', level=logging.ERROR) as log:
        client, transport = notifire_client(server, AioHTTPTransport)

        client.send_notification('foo', level='error')

        await wait(transport)

        assert server.hits[500] == 1

    msg = 'Notifire api responded with an error: APIError()'
    assert log.msgs[0] == msg


async def test_cancelled_error(fake_server, notifire_client, wait):
    server = await fake_server()

    with mock.patch('aiohttp.ClientSession.post',
                    side_effect=asyncio.CancelledError):
        client, transport = notifire_client(server, AioHTTPTransport)

        client.send_notification('foo', level='error')

        with pytest.raises(asyncio.CancelledError):
            await wait(transport)

        assert server.hits[200] == 0


async def test_async_send_when_closed(fake_server, notifire_client):
    server = await fake_server()

    with Logger('notifire', level=logging.ERROR) as log:
        client, transport = notifire_client(server, AioHTTPTransport)

        close = transport.close()

        client.send_notification('foo', level='error')

        assert server.hits[200] == 0

    assert log.msgs[0].startswith(
        'Notifire api responded with an error: AioHTTPTransport is closed')

    await close



"""
:copyright: (c) 2010-2015 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

import asyncio
from collections import defaultdict
import socket

from aiohttp import web
from aiohttp.test_utils import unused_port


class FakeResolver:
    _LOCAL_HOST = {
        0: '127.0.0.1',
        socket.AF_INET: '127.0.0.1',
        socket.AF_INET6: '::1',
    }

    def __init__(self, port):
        self.port = port

    async def resolve(self, host, port=0, family=socket.AF_INET):
        return [
            {
                'hostname': host,
                'host': self._LOCAL_HOST[family],
                'port': self.port,
                'family': family,
                'proto': 0,
                'flags': socket.AI_NUMERICHOST,
            },
        ]


class FakeServer:

    app = None
    runner = None
    site = None

    host = '127.0.0.1'

    def __init__(self, loop, side_effect=None):
        self.loop = loop

        if not side_effect:
            side_effect = {
                'status': 200,
            }

        self._side_effect = side_effect
        self._slop_factor = 0

        self.app = web.Application()
        self.setup_routes()

        self.port = unused_port()

        self.hits = defaultdict(lambda: 0)

    @property
    def side_effect(self):
        return self._side_effect

    @property
    def slop_factor(self):
        return self._slop_factor

    @slop_factor.setter
    def slop_factor(self, value):
        self._slop_factor = value

    async def start(self):
        self.runner = web.AppRunner(self.app)
        await self.runner.setup()
        self.site = web.TCPSite(self.runner, self.host, self.port)
        await self.site.start()

    def setup_routes(self):
        self.app.router.add_post('/notify', self.notify)

    async def notify(self, request):
        await asyncio.sleep(self.slop_factor, loop=self.loop)

        self.hits[self.side_effect['status']] += 1
        return web.Response(**self.side_effect)

    async def close(self):
        if self.runner:
            await self.runner.cleanup()



"""
:copyright: (c) 2010-2015 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

try:
    import requests  # NOQA, pylint: disable=W0611
except ImportError:
    import pytest
    pytest.skip(
        'Testing requests transport requires requests.',
        allow_module_level=True
    )

import mock

from notifire.client import Client
from notifire.transport import requests as uut


class TestRequestsHTTPTransport:
    client = Client('service_api_key', transport=uut.RequestsHTTPTransport)

    @mock.patch('notifire.transport.requests.requests.post')
    @mock.patch('notifire.constants.PATH')
    @mock.patch('notifire.constants.HOST')
    def test_does_send(self, host, path, post):
        host.__str__.return_value = 'localhost'
        path.__str__.return_value = 'notify'
        self.client.send_notification('foo')
        assert post.call_count == 1
        expected_url = 'localhost/notify'
        assert expected_url == post.call_args[0][0]



"""
:copyright: (c) 2010-2016 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

# Copyright (c) 2010-2013 Benjamin Peterson
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.


def check_threads():
    try:
        from uwsgi import opt
    except ImportError:
        return

    # When `threads` is passed in as an uwsgi option,
    # `enable-threads` is implied on.
    if 'threads' in opt:
        return

    if str(opt.get('enable-threads', '0')).lower() in (
            'false', 'off', 'no', '0'):
        from warnings import warn
        warn(Warning('We detected the use of uwsgi with disabled threads.  '
                     'This will cause issues with the transport you are '
                     'trying to use. Please enable threading for uwsgi.  '
                     '(Enable the "enable-threads" flag).'))



"""
:copyright: (c) 2010-2012 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

import atexit
import logging
import os
from queue import Queue
import threading
from time import sleep, time

from notifire.transport.base import AsyncTransport
from notifire.transport.http import HTTPTransport
from notifire.utils.compat import check_threads

DEFAULT_TIMEOUT = 10

logger = logging.getLogger('notifire.errors')


class AsyncWorker:
    _terminator = object()

    def __init__(self, shutdown_timeout=DEFAULT_TIMEOUT):
        check_threads()
        self._queue = Queue(-1)
        self._lock = threading.Lock()
        self._thread = None
        self._thread_for_pid = None
        self.options = {
            'shutdown_timeout': shutdown_timeout,
        }
        self.start()

    def is_alive(self):
        if self._thread_for_pid != os.getpid():
            return False
        return self._thread and self._thread.is_alive()

    def _ensure_thread(self):
        if self.is_alive():
            return
        self.start()

    def main_thread_terminated(self):
        with self._lock:
            if not self.is_alive():
                # thread not started or already stopped - nothing to do
                return

            # wake the processing thread up
            self._queue.put_nowait(self._terminator)

            timeout = self.options['shutdown_timeout']

            # wait briefly, initially
            initial_timeout = min(0.1, timeout)

            if not self._timed_queue_join(initial_timeout):
                # if that didn't work, wait a bit longer
                # NB that size is an approximation, because other threads may
                # add or remove items
                size = self._queue.qsize()

                print(
                    'Notifire is attempting to send {} pending error messages'
                    .format(size)
                )
                print('Waiting up to {} seconds'.format(timeout))

                if os.name == 'nt':
                    print('Press Ctrl-Break to quit')
                else:
                    print('Press Ctrl-C to quit')

                self._timed_queue_join(timeout - initial_timeout)

            self._thread = None

    def _timed_queue_join(self, timeout):
        """Implementation of Queue.join which takes a 'timeout' argument.

        Returns true on success, false on timeout.
        """
        deadline = time() + timeout
        queue = self._queue

        queue.all_tasks_done.acquire()
        try:
            while queue.unfinished_tasks:
                delay = deadline - time()
                if delay <= 0:
                    # timed out
                    return False

                queue.all_tasks_done.wait(timeout=delay)

            return True

        finally:
            queue.all_tasks_done.release()

    def start(self):
        """Start the task thread."""

        self._lock.acquire()
        try:
            if not self.is_alive():
                self._thread = threading.Thread(
                    target=self._target,
                    name='notifire.AsyncWorker'
                )
                self._thread.setDaemon(True)
                self._thread.start()
                self._thread_for_pid = os.getpid()
        finally:
            self._lock.release()
            atexit.register(self.main_thread_terminated)

    def stop(self, timeout=None):
        """Stop the task thread.

        Synchronous!
        """

        with self._lock:
            if self._thread:
                self._queue.put_nowait(self._terminator)
                self._thread.join(timeout=timeout)
                self._thread = None
                self._thread_for_pid = None

    def queue(self, callback, *args, **kwargs):
        self._ensure_thread()
        self._queue.put_nowait((callback, args, kwargs))

    def _target(self):
        while True:
            record = self._queue.get()
            try:
                if record is self._terminator:
                    break
                callback, args, kwargs = record
                try:
                    callback(*args, **kwargs)
                except Exception:
                    logger.error('Failed processing job', exc_info=True)
            finally:
                self._queue.task_done()

            sleep(0)


class ThreadedHTTPTransport(AsyncTransport, HTTPTransport):
    def get_worker(self):
        # pylint: disable=E0203
        if not hasattr(self, '_worker') or not self._worker.is_alive():
            self._worker = AsyncWorker()
        return self._worker

    def send_sync(self, url, data, headers, success_cb, failure_cb):
        try:
            super().send(url, data, headers)
        except Exception as e:
            failure_cb(e)
        else:
            success_cb()

    def async_send(self, url, data, headers, success_cb, failure_cb):
        self.get_worker().queue(
            self.send_sync, url, data, headers, success_cb, failure_cb)



"""
:copyright: (c) 2010-2012 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""
import urllib.request

import certifi

from notifire import constants
from notifire.exceptions import APIError, RateLimited
from notifire.transport.base import Transport
from notifire.utils.http import urlopen


class HTTPTransport(Transport):
    """The default HTTP transport."""

    def __init__(self, timeout=constants.TIMEOUT, verify_ssl=True,
                 ca_certs=None):
        self.timeout = timeout
        self.verify_ssl = verify_ssl
        self.ca_certs = ca_certs or certifi.where()

    def send(self, url, data, headers):
        """Send request to a remote webserver using HTTP POST."""

        req = urllib.request.Request(url, headers=headers)

        try:
            response = urlopen(
                url=req,
                data=data,
                timeout=self.timeout,
                verify_ssl=self.verify_ssl,
                ca_certs=self.ca_certs,
            )
        except urllib.request.HTTPError as exc:
            msg = exc.read().decode()
            code = exc.getcode()
            if code == 429:  # pylint: disable=R1720
                try:
                    retry_after = int(exc.headers.get('retry-after'))
                except (ValueError, TypeError):
                    retry_after = 0
                raise RateLimited(msg, retry_after) from exc
            elif msg:
                raise APIError(msg, code) from exc
            else:
                raise
        return response



from raven import Client as SentryClient

from notifire import Client as NotifireClient

LEVELS = {
    'critical': 'error',
    'fatal': 'error',
    'error': 'error',
    'warning': 'warning',
    'warn': 'warning',
    'info': 'info',
    'debug': 'info',
    'notset': 'info',
    50: 'error',
    40: 'error',
    30: 'warning',
    20: 'info',
    10: 'info',
    0: 'info'
}


class Client(SentryClient):
    """Client for sending notifications to Notifire via configured transport.

    Client also works independently as Sentry client
    and all of its data can be forwarded as a notification to Notifire.

    >>> from notifire import Client

    >>> # Read configuration from
    >>> # ``os.environ['NOTIFIRE_SERVICE_API_KEY']``,
    >>> # ``os.environ['SENTRY_DSN']``
    >>> client = Client()

    >>> # or pass dsn to the client
    >>> client = Client(
    >>>     service_api_key='service_api_key',
    >>>     sentry_dsn='https://public_key:secret_key@sentry.local/project_id'
    >>> )

    >>> # Send message
    >>> client.send_notification('Hello from python')

    >>> # Record an exception
    >>> try:
    >>>     1/0
    >>> except ZeroDivisionError:
    >>>     ident = client.get_ident(client.captureException())
    >>>     print "Exception caught; reference is %s" % ident
    """

    def __init__(
        self,
        service_api_key=None,
        sentry_dsn=None,
        raise_send_errors=False,
        transport=None,
        install_sys_hook=True,
        install_logging_hook=True,
        hook_libraries=None,
        enable_breadcrumbs=True,
        _random_seed=None,
        **options
    ):
        super().__init__(
            sentry_dsn,
            raise_send_errors,
            transport,
            install_sys_hook,
            install_logging_hook,
            hook_libraries,
            enable_breadcrumbs,
            _random_seed,
            **options
        )
        self.notifire_client = NotifireClient(
            service_api_key,
            raise_send_errors=raise_send_errors,
            _random_seed=_random_seed,
            **options
        )
        self.notifire_client.get_transport = self.remote.get_transport

    def send(
        self,
        auth_header=None,
        **data
    ):
        super().send(auth_header, **data)
        level = data.get('level', 'info')
        if isinstance(level, str):
            level = level.lower()
        self.send_notification(body=data['message'], level=LEVELS[level])

    def send_notification(
        self,
        body,
        level='info',
        text=None,
        url=None,
        sample_rate=None
    ):
        self.notifire_client.send_notification(
            body,
            level,
            text,
            url,
            sample_rate
        )



import mock
import pytest

from notifire import client as uut
from notifire.message import Message


class TestClient:
    client = uut.Client('service_api_key')

    @mock.patch('notifire.constants.CLIENT')
    @mock.patch('notifire.dsn.datetime')
    def test_build_headers(self, time, notifire_client):
        time.utcnow().__sub__().total_seconds.return_value = '123456.789'
        notifire_client.__str__.return_value = 'notifire-python'

        assert self.client._build_headers() == {
            'Content-Type': 'application/json',
            'Content-Encoding': 'gzip',
            'X-Notifire-Auth': (
                'Notifire notifire_api_key=service_api_key, '
                'notifire_timestamp=123456.789, '
                'notifire_client={}'.format('notifire-python')
            )

        }

    @mock.patch('notifire.client.Client._send')
    @mock.patch('notifire.constants.CLIENT')
    @mock.patch('notifire.constants.PATH')
    @mock.patch('notifire.constants.HOST')
    @mock.patch('notifire.dsn.datetime')
    def test_sendNotification(self, time, host, path, notifire_client, _send):
        time.utcnow().__sub__().total_seconds.return_value = '123456.789'
        notifire_client.__str__.return_value = 'notifire-python'
        host.__str__.return_value = 'localhost'
        path.__str__.return_value = 'notify'

        self.client.send_notification(body='foo', text='bar')
        _send.assert_called_once_with(
            url='localhost/notify',
            data=Message.get_payload(body='foo', text='bar'),
            headers={
                'Content-Type': 'application/json',
                'Content-Encoding': 'gzip',
                'X-Notifire-Auth': (
                    'Notifire notifire_api_key=service_api_key, '
                    'notifire_timestamp=123456.789, '
                    'notifire_client={}'.format('notifire-python')
                )
            }
        )

        self.client.send_notification(body='foo', text='bar', sample_rate=0)
        _send.assert_called_once()

    @mock.patch('notifire.client.Client._send')
    @mock.patch('notifire.constants.CLIENT')
    @mock.patch('notifire.constants.PATH')
    @mock.patch('notifire.constants.HOST')
    @mock.patch('notifire.dsn.datetime')
    def test_send(self, time, host, path, notifire_client, _send):
        time.utcnow().__sub__().total_seconds.return_value = '123456.789'
        notifire_client.__str__.return_value = 'notifire-python'
        host.__str__.return_value = 'localhost'
        path.__str__.return_value = 'notify'

        data = Message.get_payload(body='foo', text='bar')
        self.client.send(data)

        _send.assert_called_once_with(
            url='localhost/notify',
            data=data,
            headers={
                'Content-Type': 'application/json',
                'Content-Encoding': 'gzip',
                'X-Notifire-Auth': (
                    'Notifire notifire_api_key=service_api_key, '
                    'notifire_timestamp=123456.789, '
                    'notifire_client={}'.format('notifire-python')
                )
            }
        )

        _send.side_effect = Exception()
        with pytest.raises(Exception):
            self.client.send(data)



import logging
import os
from random import Random

from notifire.dsn import Dsn
from notifire.exceptions import APIError
from notifire.message import Message
from notifire.transport.threaded import ThreadedHTTPTransport

log = logging.getLogger('notifire')


class Client:
    """Client for sending notifications to Notifire via configured transport.

    >>> from notifire import Client

    >>> # Read service api key configuration from
    >>> # ``os.environ['NOTIFIRE_SERVICE_API_KEY']``
    >>> client = Client()

    >>> # or pass service api key to the client
    >>> client = Client(service_api_key='service_api_key')

    >>> # Send message
    >>> client.send_notification('Hello from Python')
    """

    def __init__(
        self,
        service_api_key=None,
        transport=None,
        raise_send_errors=False,
        _random_seed=None,
        **options
    ):
        service_api_key = os.environ.get('NOTIFIRE_SERVICE_API_KEY',
                                         service_api_key)
        self.dsn = Dsn(service_api_key)
        self._transport = transport or ThreadedHTTPTransport
        self._random = Random(_random_seed)
        self.raise_send_errors = raise_send_errors
        self.sample_rate = options.get('sample_rate', 1)
        self.success_callback = options.get('success_callback',
                                            self.successful_send)

    def get_transport(self):
        """Instantiate transport class outside of init.

        This prevents creation of loop or spawning new thread too
        early. Also ensuring that only one instance of transport class
        will be present.
        """
        if not hasattr(self, 'transport'):
            self.transport = self._transport()
        return self.transport

    def send_notification(self, body, level='info', text=None, url=None,
                          sample_rate=None):
        """Capture and send message to notifire application.

        :param body: str, the title of the iOS notification pop-up
        :param level: str, one of ('info', 'warning', 'error')
                      to filter the notification
        :param text: str, additional text of the notification
                     shown in notification tab
        :param url: str, url displayed in the notification
        :param sample_rate: float, in the range of [0, 1], process rate
                            of this event
        :return:
        """
        if sample_rate is None:
            sample_rate = self.sample_rate

        if self._random.random() >= sample_rate:
            return

        data = Message.get_payload(body, level, text, url)
        self.send(data)

    def _build_headers(self):
        return {
            'X-Notifire-Auth': self.dsn.auth_header(),
            'Content-Type': 'application/json',
            'Content-Encoding': 'gzip'
        }

    def send(self, data):
        return self._send(
            url=self.dsn.api_url,
            data=data,
            headers=self._build_headers()
        )

    def _send(self, url, data, headers):
        def failed_send(e):
            self._failed_send(e, url, Message.decode(data))

        try:
            transport = self.get_transport()
            if transport.is_async:
                transport.async_send(url, data, headers, self.successful_send,
                                     failed_send)
            else:
                transport.send(url, data, headers)
        except Exception as e:
            if self.raise_send_errors:
                raise
            failed_send(e)

    @staticmethod
    def _failed_send(exc, url, data):
        if isinstance(exc, APIError):
            log.error('Notifire api responded with an error: %s(%s)',
                      type(exc).__name__, exc.message)

        else:
            log.error('Notifire api responded with an error: %s (url: %s)',
                      exc, url, exc_info=True, extra={'data': data})

    @staticmethod
    def successful_send():
        """Callback for successful message delivery."""
        return



import mock
import pytest

from notifire import cli as uut


@mock.patch('notifire.client.Client.send_notification')
def test_main(client):
    kwargs = {
        'body': 'foo',
        'level': 'error',
        'text': 'bar',
        'url': 'https://notifire.notifire'
    }
    with mock.patch(
        'argparse._sys.argv', [
            'notifire-cli',
            '--service_api_key',
            'service_api_key',
            '--body',
            kwargs['body'],
            '--level',
            kwargs['level'],
            '--text',
            kwargs['text'],
            '--url',
            kwargs['url']
        ]
    ):
        uut.main()
    client.assert_called_with(**kwargs)

    with mock.patch(
        'argparse._sys.argv', [
            'notifire-cli',
            '--service_api',
            'service_api_key',
            '--body',
            kwargs['body'],
            '--random_arg',
            'this should trigger error'
        ]
    ):
        with pytest.raises(SystemExit):
            uut.main()



import mock

from notifire import constants
from notifire import dsn as uut


class TestDsn:
    service_api_key = 'service_api_key'
    dsn = uut.Dsn(service_api_key)

    def test_init(self):
        assert self.dsn.service_api_key == self.service_api_key

    @mock.patch('notifire.constants.PATH')
    @mock.patch('notifire.constants.HOST')
    def test_api_url(self, host, path):
        host.__str__.return_value = 'localhost'
        path.__str__.return_value = 'notify'

        assert self.dsn.api_url == 'localhost/notify'

    def test_auth_header(self):
        with mock.patch('notifire.dsn.datetime') as time:
            time.utcnow().__sub__().total_seconds.return_value = '123456.789'
            assert self.dsn.auth_header() == (
                'Notifire notifire_api_key={0}, notifire_timestamp={1}, '
                'notifire_client={2}'
                .format(self.service_api_key, '123456.789', constants.CLIENT)
            )



"""
:copyright: (c) 2010-2012 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from notifire.exceptions import APIError, RateLimited
from notifire.transport.http import HTTPTransport

try:
    import requests
    has_requests = True
except ImportError:
    has_requests = False


class RequestsHTTPTransport(HTTPTransport):
    def __init__(self, *args, **kwargs):
        if not has_requests:
            raise ImportError('RequestsHTTPTransport requires requests.')

        super().__init__(*args, **kwargs)

    def send(self, url, data, headers):
        if self.verify_ssl:
            # If SSL verification is enabled, use the provided CA bundle to
            # perform the verification
            self.verify_ssl = self.ca_certs
        try:
            resp = requests.post(url, data=data, headers=headers,
                                 verify=self.verify_ssl, timeout=self.timeout)
            resp.raise_for_status()
        except requests.exceptions.HTTPError as exc:
            msg = exc.response.text
            code = exc.response.status_code
            print(exc.response.headers.get('retry-after'))
            if code == 429:  # pylint: disable=R1720
                try:
                    retry_after = exc.response.headers.get('retry-after')
                    retry_after = int(retry_after)
                except (ValueError, TypeError):
                    retry_after = 0
                raise RateLimited(msg, retry_after) from exc
            elif msg:
                raise APIError(msg, code) from exc
            else:
                raise



from datetime import datetime

from notifire import constants

EPOCH = datetime(1970, 1, 1)


class Dsn:
    def __init__(self, service_api_key):
        self.service_api_key = service_api_key

    @property
    def api_url(self):
        return '{host}/{path}'.format(
            host=constants.HOST,
            path=constants.PATH
        )

    def auth_header(self):
        header = [
            ('notifire_api_key', self.service_api_key),
            ('notifire_timestamp', (datetime.utcnow() - EPOCH)
             .total_seconds()),
            ('notifire_client', constants.CLIENT)
        ]
        return 'Notifire ' + ', '.join('{}={}'.format(key, value)
                                       for key, value in header)



"""
:copyright: (c) 2010-2012 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""


class Transport:
    """All transport implementations need to subclass this class.

    You must implement a send method (or an async_send method if
    sub-classing AsyncTransport).
    """

    is_async = False

    def send(self, url, data, headers):
        """You need to override this to do something with the actual data.

        Usually - this is sending to a server.
        """
        raise NotImplementedError


class AsyncTransport(Transport):
    """All asynchronous transport implementations should subclass this class.

    You must implement an async_send method.
    """

    is_async = True

    def async_send(self, url, data, headers, success_cb, error_cb):
        """Override this method for asynchronous transports.

        Call `success_cb()` if the send succeeds or `error_cb(exception)`
        if the send fails.
        """
        raise NotImplementedError



"""
:copyright: (c) 2010-2015 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

import asyncio

import pytest

from notifire.transport.aiohttp import AioHTTPTransport


def test_loop_is_none():
    with pytest.raises(RuntimeError):
        AioHTTPTransport()


def test_explicit_loop(event_loop):
    _transport = AioHTTPTransport(loop=event_loop)

    assert _transport._loop == event_loop

    event_loop.run_until_complete(_transport.close())


def test_global_loop(event_loop):
    asyncio.set_event_loop(event_loop)

    _transport = AioHTTPTransport()

    assert _transport._loop == event_loop

    event_loop.run_until_complete(_transport.close())


def test_transport_closed_twice(event_loop, mocker):
    _transport = AioHTTPTransport(loop=event_loop)

    event_loop.run_until_complete(_transport.close())

    close_coro = mocker.spy(_transport, '_close_coro')

    event_loop.run_until_complete(_transport.close())

    assert close_coro.call_count == 0



"""
:copyright: (c) 2010-2015 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

import logging


class TestHandler(logging.Handler):
    def __init__(self, records, msgs, *args, **kwargs):
        self.records = records
        self.msgs = msgs
        super().__init__(*args, **kwargs)

    def emit(self, record):
        self.records.append(record)

        self.msgs.append(self.format(record))


class Logger:
    def __init__(self, name, level=logging.NOTSET):
        self.name = name
        self.level = level

        self.records, self.msgs = [], []

        self.logger = logging.getLogger(self.name)
        self.handler = TestHandler(self.records, self.msgs, level=self.level)

    def __enter__(self):
        self.logger.addHandler(self.handler)

        return self

    def __exit__(self, *exc_info):
        self.logger.removeHandler(self.handler)



"""
:copyright: (c) 2010-2015 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

import asyncio
from functools import partial
import gc
import os

import async_timeout
from mock import patch
import pytest

from notifire.client import Client
from tests.transport.test_aiohttp.fake_server import FakeResolver, FakeServer

asyncio.set_event_loop(None)


@pytest.fixture
def event_loop(request):
    asyncio.set_event_loop(None)
    loop = asyncio.new_event_loop()
    loop.set_debug(bool(os.environ.get('PYTHONASYNCIODEBUG')))
    request.addfinalizer(lambda: asyncio.set_event_loop(None))

    yield loop

    loop.call_soon(loop.stop)
    loop.run_forever()
    loop.close()

    gc.collect()


@pytest.fixture
def notifire_client(event_loop):
    transports = []

    def do_client(fake_server, cls, *args, **kwargs):
        kwargs.setdefault('loop', event_loop)

        service_api_key = 'service_api_key'
        host = patch(
            'notifire.constants.HOST', 'http://127.0.0.1:{port}'.format(
                port=fake_server.port
            )
        )
        host.start()
        client = Client(
            service_api_key,
            transport=partial(cls, *args, **kwargs)
        )

        transport = client.get_transport()
        resolver = FakeResolver(fake_server.port)
        transport._client_session._connector._resolver = resolver

        transports.append(transport)

        return client, transport

    yield do_client

    async def do_close():
        closes = [transport.close() for transport in transports]
        await asyncio.gather(*closes, loop=event_loop)

    event_loop.run_until_complete(do_close())


@pytest.fixture
def fake_server(event_loop):
    servers = []

    async def do_server(*args, **kwargs):
        kwargs.setdefault('loop', event_loop)
        server = FakeServer(*args, **kwargs)
        servers.append(server)

        await server.start()

        return server

    yield do_server

    async def do_close():
        closes = [server.close() for server in servers]
        await asyncio.gather(*closes, loop=event_loop)

    event_loop.run_until_complete(do_close())


@pytest.fixture
def wait(event_loop):
    async def do_wait(transport, timeout=1):
        coro = asyncio.gather(*transport._tasks, loop=event_loop)

        async with async_timeout.timeout(timeout, loop=event_loop):
            await coro

    return do_wait



from setuptools import find_packages, setup

install_requires = ['certifi']
raven_requires = [
    'raven',
]
tests_require = [
    'aiohttp',
    'gevent'
    'mock',
    'pytest',
    'pytest-asyncio',
    'pytest-mock',
    'requests'
] + raven_requires

setup(
    name='notifire',
    version='0.1.1',
    url='https://github.com/Notifire/notifire-python',
    author='Notifire team',
    author_email='notifire@memonil.com',
    description='Python client for Notifire (https://notifire.dvdblk.com)',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    packages=find_packages(exclude=('tests', 'tests.*')),
    license='BSD',
    install_requires=install_requires,
    extras_require={
        'raven': raven_requires,
    },
    tests_require=tests_require,
    entry_points={
        'console_scripts': ['notifire-cli=notifire.cli:main']
    },
    classifiers=[
        'Intended Audience :: Developers',
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



"""
:copyright: (c) 2010-2015 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

try:
    import requests  # NOQA, pylint: disable=W0611
except ImportError:
    import pytest
    pytest.skip(
        'Testing requests transport requires requests.',
        allow_module_level=True
    )

import time

import mock

from notifire.client import Client
from notifire.message import Message
from notifire.transport import threaded_requests as uut


class DummyThreadedScheme(uut.ThreadedRequestsHTTPTransport):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.events = []
        self.send_delay = 0

    def send_sync(self, url, data, headers, success_cb, failure_cb):
        # delay sending the message, to allow us to tests that the shutdown
        # hook waits correctly
        time.sleep(self.send_delay)

        self.events.append((url, data, headers, success_cb, failure_cb))


class TestThreadedRequestsHTTPTransport:
    message = Message()
    client = Client(
        'service_api_key',
        transport=uut.ThreadedRequestsHTTPTransport
    )

    @mock.patch('notifire.transport.requests.requests.post')
    @mock.patch('notifire.constants.PATH')
    @mock.patch('notifire.constants.HOST')
    def test_does_send(self, host, path, send):
        host.__str__.return_value = 'localhost'
        path.__str__.return_value = 'notify'
        self.client.send_notification('foo')

        time.sleep(0.1)

        assert send.call_count == 1
        expected_url = 'localhost/notify'
        assert expected_url == send.call_args[0][0]

    def test_shutdown_waits_for_send(self):
        transport = DummyThreadedScheme()
        transport.send_delay = 0.5

        data = self.message.get_payload(
            'foo',
            level='info',
            text=None,
            url=None
        )
        transport.async_send('localhost', data, None, None, None)

        time.sleep(0.1)

        # this should wait for the message to get sent
        transport.get_worker().main_thread_terminated()

        assert len(transport.events) == 1



"""
:copyright: (c) 2010-2012 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from http import client
import socket
import ssl
import urllib.request

from notifire import constants
from notifire.utils.ssl_match_hostname import match_hostname


def urlopen(url, data=None, timeout=constants.TIMEOUT, ca_certs=None,
            verify_ssl=False, assert_hostname=None):

    class ValidHTTPSConnection(client.HTTPConnection):
        default_port = client.HTTPS_PORT

        def __init__(self, *args, **kwargs):
            client.HTTPConnection.__init__(self, *args, **kwargs)

        def connect(self):
            sock = socket.create_connection(
                address=(self.host, self.port),
                timeout=self.timeout,
            )
            if self._tunnel_host:
                self.sock = sock
                self._tunnel()

            self.sock = ssl.wrap_socket(
                sock, ca_certs=ca_certs, cert_reqs=ssl.CERT_REQUIRED)

            if assert_hostname is not None:
                match_hostname(self.sock.getpeercert(),
                               assert_hostname or self.host)

    class ValidHTTPSHandler(urllib.request.HTTPSHandler):
        def https_open(self, req):
            return self.do_open(ValidHTTPSConnection, req)

    if verify_ssl:
        handlers = [ValidHTTPSHandler]
    else:
        try:
            handlers = [urllib.request.HTTPSHandler(
                context=ssl._create_unverified_context())]
        except AttributeError:
            handlers = []

    opener = urllib.request.build_opener(*handlers)
    return opener.open(url, data, timeout)



import mock

from notifire.client import Client
from notifire.transport import http as uut


class TestRequestsHTTPTransport:
    client = Client('service_api_key', transport=uut.HTTPTransport)

    @mock.patch('urllib.request.OpenerDirector.open')
    def test_does_send(self, request):
        self.client.send_notification('foo')
        assert request.call_count == 1



import argparse

from notifire.client import Client
from notifire.transport import HTTPTransport


def main():
    parser = argparse.ArgumentParser(
        prog='notifire',
        description='cli for notifire client library')
    parser.add_argument(
        '-s',
        '--service_api_key',
        required=True,
        help='service api key')
    parser.add_argument(
        '-b',
        '--body',
        required=True,
        help='the title of the iOS notification pop-up')
    parser.add_argument(
        '-l',
        '--level',
        default='info',
        help="one of ('info', 'warning', 'error') to filter the notification")
    parser.add_argument(
        '-t',
        '--text',
        help='additional text of the notification shown in notification tab')
    parser.add_argument(
        '-u',
        '--url',
        help='url displayed in the notification')
    args = vars(parser.parse_args())
    service_api_key = args.pop('service_api_key')
    Client(service_api_key, transport=HTTPTransport).send_notification(**args)



try:
    from raven import Client as SentryClient
except ImportError:
    import pytest
    pytest.skip(
        'Testing raven transport requires raven.',
        allow_module_level=True
    )

import mock

from notifire import Client as NotifireClient
from notifire.integrations import raven as uut
from notifire.transport.base import AsyncTransport


class TestClient:
    service_api_key = 'service_api_key'
    sentry_dsn = 'https://public_key:secret_key@sentry.local/project_id'
    client = uut.Client(
        service_api_key=service_api_key,
        sentry_dsn=sentry_dsn
    )

    def test_init(self):
        assert isinstance(self.client, SentryClient)
        assert isinstance(self.client.notifire_client, NotifireClient)

    @staticmethod
    @mock.patch('raven.conf.remote.RemoteConfig.get_transport')
    def test_shared_transport(get_transport):
        transport = mock.Mock(spec=AsyncTransport)
        get_transport.return_value = transport
        # not using already created client instance because of patching
        # that affects init
        service_api_key = 'service_api_key'
        sentry_dsn = 'https://public_key:secret_key@sentry.local/project_id'
        client = uut.Client(
            service_api_key=service_api_key,
            sentry_dsn=sentry_dsn
        )

        try:
            1 / 0
        except ZeroDivisionError:
            client.captureException()

        # request should be sent over the same transport instance
        assert transport.async_send.call_count == 2
        client.send_notification(body='foo', text='bar')
        assert transport.async_send.call_count == 3

    @mock.patch('raven.base.Client.send')
    @mock.patch('raven.base.Client.build_msg')
    @mock.patch('notifire.client.Client.send_notification')
    def test_send(self, sendNotification, build_msg, send):
        data = {
            'level': 40,
            'message': 'ZeroDivisionError: division by zero',
            'event_id': '1337'
        }
        build_msg.return_value = data

        try:
            1 / 0
        except ZeroDivisionError:
            self.client.captureException()

        send.assert_called_once_with(None, **data)
        sendNotification.assert_called_once_with(
            data['message'],
            'error',
            None, None, None
        )

    @mock.patch('raven.base.Client.send_remote')
    @mock.patch('notifire.client.Client.send_notification')
    def test_sendNotification(self, sendNotification, send_remote):
        self.client.send_notification(body='foo', text='bar')
        send_remote.assert_not_called()
        sendNotification.assert_called_once_with(
            'foo',
            'info',
            'bar',
            None,
            None
        )
